/*
Author: Cao Thanh Tung
Date: 01/03/2009

File Name: cudaMalloc.cu

This file include all CUDA code to perform some simple GPU memory management.

===============================================================================

Copyright (c) 2009, School of Computing, National University of Singapore. 
All rights reserved.

Project homepage: http://www.comp.nus.edu.sg/~tants/delaunay.html

If you use GPUDT and you like it or have comments on its usefulness etc., we 
would love to hear from you at <tants@comp.nus.edu.sg>. You may share with us
your experience and any possibilities that we may improve the work/code.

===============================================================================

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of
conditions and the following disclaimer. Redistributions in binary form must reproduce
the above copyright notice, this list of conditions and the following disclaimer
in the documentation and/or other materials provided with the distribution. 

Neither the name of the National University of University nor the names of its contributors
may be used to endorse or promote products derived from this software without specific
prior written permission from the National University of Singapore. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO THE IMPLIED WARRANTIES 
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

*/

#include "cuda.h"
#include "common.h"

#include <stdio.h>

#ifdef SINGLE_PRECISION
#define MEM_BLOCK	32
#else
#define MEM_BLOCK	40
#endif

int *memAvailable[MEM_BLOCK]; 
void *memPool;
int memBlockSize; 

void cudaMemInitialize(int blockSize) 
{
	// Memory alignment
	memBlockSize = (blockSize / 256 + 1) * 256; 

	// Allocate the memory pool
	cutilSafeCall( cudaMalloc(&memPool, memBlockSize * MEM_BLOCK) ); 

	for (int i = 0; i < MEM_BLOCK; i++)
		memAvailable[i] = NULL; 
}

int *cudaMemAlloc(int nBlock) 
{
	// Find an available slot
	int count = 0, i = 0; 

	while (count < nBlock && count < MEM_BLOCK) {
		if (memAvailable[i] == NULL)
			count++; 
		else
			count = 0; 

		i++; 
	}

	if (count >= MEM_BLOCK) 
	{
		printf("cudaMemAllocation: Not enough blocks.\n"); 
		exit(-1); 
	}

	// Here's the found location
	i -= nBlock; 

	// Mark as used. 
	int *ptr = (int *) ((char *) memPool + (i * memBlockSize)); 

	for (int j = 0; j < nBlock; j++)
		memAvailable[i + j] = ptr; 

	return ptr; 
}

void cudaMemFree(void *ptr) 
{
	for (int i = 0; i < MEM_BLOCK; i++)
		if (memAvailable[i] == ptr) {
			memAvailable[i] = NULL; 
		}
}

void cudaMemDeallocation() 
{
	cutilSafeCall( cudaFree(memPool) ); 
}